"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const url_1 = require("url");
const link_1 = require("./link");
const query_1 = require("./query");
const types_1 = require("./types");
const validators_1 = require("./validators");
class BugzillaAPI {
    constructor(instance, user, password, restrictLogin = false) {
        let url = instance instanceof url_1.URL ? instance : new url_1.URL(instance);
        if (!user) {
            this.link = new link_1.PublicLink(url);
        }
        else if (password !== undefined) {
            this.link = new link_1.PasswordLink(url, user, password, restrictLogin);
        }
        else {
            this.link = new link_1.ApiKeyLink(url, user);
        }
    }
    async version() {
        let version = await this.link.get('version', (0, validators_1.object)(types_1.VersionSpec));
        return version.version;
    }
    whoami() {
        return this.link.get('whoami', (0, validators_1.object)(types_1.UserSpec));
    }
    async bugHistory(bugId, since) {
        var _a;
        let searchParams;
        if (since) {
            searchParams = new url_1.URLSearchParams();
            searchParams.set('new_since', (_a = since.toISODate()) !== null && _a !== void 0 ? _a : '');
        }
        let bugs = await this.link.get(`bug/${bugId}/history`, (0, validators_1.object)(types_1.HistoryLookupSpec), searchParams);
        let [bug] = bugs.bugs;
        if (!bug) {
            throw new Error('Bug not found.');
        }
        return bug.history;
    }
    searchBugs(query) {
        return new query_1.FilteredQuery(async (includes, excludes) => {
            let search = (0, link_1.params)(query);
            if (includes) {
                search.set('include_fields', includes.join(','));
            }
            if (excludes) {
                search.set('exclude_fields', excludes.join(','));
            }
            let result = await this.link.get('bug', (0, validators_1.object)({
                bugs: (0, validators_1.array)((0, validators_1.object)(types_1.BugSpec, includes, excludes)),
            }), search);
            return result.bugs;
        });
    }
    getBugs(ids) {
        return this.searchBugs({
            id: ids.join(','),
        });
    }
    quicksearch(query) {
        return this.searchBugs({
            quicksearch: query,
        });
    }
    advancedSearch(query) {
        let searchParams;
        if (query instanceof url_1.URL) {
            searchParams = query.searchParams;
        }
        else if (typeof query == 'string' && query.startsWith('http')) {
            searchParams = new url_1.URL(query).searchParams;
        }
        else if (query instanceof url_1.URLSearchParams) {
            searchParams = query;
        }
        else {
            searchParams = new url_1.URLSearchParams(query);
        }
        searchParams.delete('list_id');
        return this.searchBugs(searchParams);
    }
    async getComment(commentId) {
        let comment = await this.link.get(`bug/comment/${commentId}`, (0, validators_1.object)(types_1.CommentsSpec));
        if (!comment) {
            throw new Error(`Failed to get comment #${commentId}.`);
        }
        return comment.comments.get(commentId);
    }
    async getBugComments(bugId) {
        var _a;
        let comments = await this.link.get(`bug/${bugId}/comment`, (0, validators_1.object)(types_1.CommentsSpec));
        if (!comments) {
            throw new Error(`Failed to get comments of bug #${bugId}.`);
        }
        return (_a = comments.bugs.get(bugId)) === null || _a === void 0 ? void 0 : _a.comments;
    }
    async createComment(bugId, comment, options = {}) {
        var _a;
        const content = {
            comment,
            is_private: (_a = options.is_private) !== null && _a !== void 0 ? _a : false,
        };
        let commentStatus = await this.link.post(`bug/${bugId}/comment`, (0, validators_1.object)(types_1.CreatedCommentSpec), content);
        if (!commentStatus) {
            throw new Error('Failed to create comment.');
        }
        return commentStatus.id;
    }
    async createBug(bug) {
        let bugStatus = await this.link.post('bug', (0, validators_1.object)(types_1.CreatedBugSpec), bug);
        if (!bugStatus) {
            throw new Error('Failed to create bug.');
        }
        return bugStatus.id;
    }
    async updateBug(bugIdOrAlias, data) {
        let response = await this.link.put(`bug/${bugIdOrAlias}`, (0, validators_1.object)(types_1.UpdatedBugTemplateSpec), data);
        if (!response) {
            throw new Error(`Failed to update bug #${bugIdOrAlias}.`);
        }
        return response.bugs;
    }
    async getAttachment(attachmentId) {
        let attachment = await this.link.get(`bug/attachment/${attachmentId}`, (0, validators_1.object)(types_1.AttachmentsSpec));
        if (!attachment) {
            throw new Error(`Failed to get attachment #${attachmentId}.`);
        }
        return attachment.attachments.get(attachmentId);
    }
    async getBugAttachments(bugId) {
        let attachments = await this.link.get(`bug/${bugId}/attachment`, (0, validators_1.object)(types_1.AttachmentsSpec));
        if (!attachments) {
            throw new Error(`Failed to get attachments of bug #${bugId}.`);
        }
        return attachments.bugs.get(bugId);
    }
    async createAttachment(bugId, attachment) {
        const dataBase64 = {
            data: Buffer.from(attachment.data).toString('base64'),
        };
        let attachmentStatus = await this.link.post(`bug/${bugId}/attachment`, (0, validators_1.object)(types_1.CreatedAttachmentSpec), { ...attachment, ...dataBase64 });
        if (!attachmentStatus) {
            throw new Error('Failed to create attachment.');
        }
        return attachmentStatus.ids;
    }
    async updateAttachment(attachmentId, data) {
        let response = await this.link.put(`bug/attachment/${attachmentId}`, (0, validators_1.object)(types_1.UpdatedAttachmentTemplateSpec), data);
        if (!response) {
            throw new Error(`Failed to update attachment #${attachmentId}.`);
        }
        return response.attachments;
    }
}
exports.default = BugzillaAPI;
//# sourceMappingURL=index.js.map