"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PasswordLink = exports.ApiKeyLink = exports.PublicLink = exports.BugzillaLink = exports.params = void 0;
const url_1 = require("url");
const axios_1 = __importDefault(require("axios"));
const validators_1 = require("./validators");
const types_1 = require("./types");
function params(search) {
    if (search instanceof url_1.URLSearchParams) {
        return search;
    }
    return new url_1.URLSearchParams(search);
}
exports.params = params;
function isError(payload) {
    // @ts-ignore
    // eslint-disable-next-line @typescript-eslint/no-unsafe-return
    return payload && typeof payload == 'object' && payload.error;
}
async function performRequest(config, validator) {
    var _a;
    try {
        let response = await axios_1.default.request({
            ...config,
            headers: {
                Accept: 'application/json',
                ...((_a = config.headers) !== null && _a !== void 0 ? _a : {}),
            },
        });
        if (isError(response.data)) {
            throw new Error(response.data.message);
        }
        return validator(response.data);
    }
    catch (e) {
        if (axios_1.default.isAxiosError(e)) {
            throw new Error(e.message);
        }
        else {
            throw e;
        }
    }
}
/**
 * Responsible for requesting data from the bugzilla instance handling any
 * necessary authentication and error handling that must happen. The chief
 * access is through the `get`, `post` and `put` methods.
 */
class BugzillaLink {
    constructor(instance) {
        this.instance = new url_1.URL('rest/', instance);
    }
    buildURL(path, query) {
        let url = new url_1.URL(path, this.instance);
        if (query) {
            url.search = params(query).toString();
        }
        return url;
    }
    async get(path, validator, searchParams) {
        return this.request({
            url: this.buildURL(path, searchParams).toString(),
        }, validator);
    }
    async post(path, validator, content, searchParams) {
        return this.request({
            url: this.buildURL(path, searchParams).toString(),
            method: 'POST',
            data: JSON.stringify(content),
            headers: {
                'Content-Type': 'application/json',
            },
        }, validator);
    }
    async put(path, validator, content, searchParams) {
        return this.request({
            url: this.buildURL(path, searchParams).toString(),
            method: 'PUT',
            data: JSON.stringify(content),
            headers: {
                'Content-Type': 'application/json',
            },
        }, validator);
    }
}
exports.BugzillaLink = BugzillaLink;
class PublicLink extends BugzillaLink {
    async request(config, validator) {
        return performRequest(config, validator);
    }
}
exports.PublicLink = PublicLink;
/**
 * Handles authentication using an API key.
 */
class ApiKeyLink extends BugzillaLink {
    constructor(instance, apiKey) {
        super(instance);
        this.apiKey = apiKey;
    }
    async request(config, validator) {
        var _a;
        return performRequest({
            ...config,
            headers: {
                ...((_a = config.headers) !== null && _a !== void 0 ? _a : {}),
                'X-BUGZILLA-API-KEY': this.apiKey,
                // Red Hat Bugzilla uses Authorization header - https://bugzilla.redhat.com/docs/en/html/api/core/v1/general.html#authentication
                Authorization: `Bearer ${this.apiKey}`,
            },
        }, validator);
    }
}
exports.ApiKeyLink = ApiKeyLink;
/**
 * Handles authentication using a username and password.
 */
class PasswordLink extends BugzillaLink {
    constructor(instance, username, password, restrictLogin) {
        super(instance);
        this.username = username;
        this.password = password;
        this.restrictLogin = restrictLogin;
        this.token = null;
    }
    async login() {
        let loginInfo = await performRequest({
            url: this.buildURL('login', {
                login: this.username,
                password: this.password,
                restrict_login: String(this.restrictLogin),
            }).toString(),
        }, (0, validators_1.object)(types_1.LoginResponseSpec));
        return loginInfo.token;
    }
    async request(config, validator) {
        var _a;
        if (!this.token) {
            this.token = await this.login();
        }
        return performRequest({
            ...config,
            headers: {
                ...((_a = config.headers) !== null && _a !== void 0 ? _a : {}),
                'X-BUGZILLA-TOKEN': this.token,
            },
        }, validator);
    }
}
exports.PasswordLink = PasswordLink;
//# sourceMappingURL=link.js.map