"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.either = exports.map = exports.intString = exports.maybeArray = exports.optional = exports.nullable = exports.base64 = exports.string = exports.double = exports.int = exports.boolean = exports.datetime = exports.array = exports.object = void 0;
const luxon_1 = require("luxon");
function repr(val) {
    return `\`${JSON.stringify(val)}\``;
}
function object(validator, includes = Object.keys(validator), excludes = []) {
    return (val) => {
        if (!val || typeof val != 'object') {
            throw new Error(`Expected an object but received ${repr(val)}`);
        }
        let result = {};
        for (let field of includes) {
            if (excludes.includes(field)) {
                continue;
            }
            let fieldValidator = validator[field];
            if (!fieldValidator) {
                continue;
            }
            try {
                result[field] = fieldValidator(val[field]);
            }
            catch (e) {
                throw new Error(`Error validating field '${field}': ${e.message}`);
            }
        }
        return result;
    };
}
exports.object = object;
function array(validator) {
    return (val) => {
        if (!Array.isArray(val)) {
            throw new Error(`Expected an array but received ${repr(val)}`);
        }
        try {
            return val.map(validator);
        }
        catch (e) {
            throw new Error(`Error validating array: ${e.message}`);
        }
    };
}
exports.array = array;
function typedValidator(type) {
    return (val) => {
        if (val === null || typeof val != type) {
            throw new Error(`Expected a ${type} but received ${repr(val)}`);
        }
        return val;
    };
}
function datetime(val) {
    if (typeof val != 'string') {
        throw new Error(`Expected an ISO-8601 string but received ${repr(val)}`);
    }
    let dt = luxon_1.DateTime.fromISO(val);
    if (!dt.isValid) {
        throw new Error(`Expected an ISO-8601 string but received ${repr(val)}`);
    }
    return dt;
}
exports.datetime = datetime;
exports.boolean = typedValidator('boolean');
exports.int = typedValidator('number');
exports.double = typedValidator('number');
exports.string = typedValidator('string');
function base64(val) {
    if (typeof val != 'string') {
        throw new Error(`Expected a base64 encoded string but received ${repr(val)}`);
    }
    try {
        return Buffer.from(val, 'base64');
    }
    catch (_) {
        throw new Error(`Expected a base64 encoded string but received ${repr(val)}`);
    }
}
exports.base64 = base64;
function nullable(validator, result) {
    return (val) => {
        if (val === null) {
            return result !== null && result !== void 0 ? result : null;
        }
        return validator(val);
    };
}
exports.nullable = nullable;
function optional(validator, result) {
    return (val) => {
        if (val === undefined) {
            return result;
        }
        return validator(val);
    };
}
exports.optional = optional;
function maybeArray(validator) {
    return (val) => {
        if (Array.isArray(val)) {
            try {
                return val.map(validator);
            }
            catch (e) {
                throw new Error(`Error validating array: ${e.message}`);
            }
        }
        return validator(val);
    };
}
exports.maybeArray = maybeArray;
function intString(val) {
    if (typeof val != 'string') {
        throw new Error(`Expected an integer as a string but received ${repr(val)}`);
    }
    let value = parseInt(val, 10);
    if (value.toString() != val) {
        throw new Error(`Expected an integer as a string but received ${repr(val)}`);
    }
    return value;
}
exports.intString = intString;
function map(keyValidator, valueValidator) {
    return (val) => {
        if (!val || typeof val != 'object') {
            throw new Error(`Expected an object but received ${repr(val)}`);
        }
        let result = new Map();
        for (let [key, value] of Object.entries(val)) {
            result.set(keyValidator(key), valueValidator(value));
        }
        return result;
    };
}
exports.map = map;
function either(first, second) {
    return (val) => {
        let result;
        try {
            result = first(val);
        }
        catch (_) {
            try {
                result = second(val);
            }
            catch (e) {
                throw new Error(`Expected an ${first.name} or ${second.name} but received ${repr(val)}`);
            }
        }
        return result;
    };
}
exports.either = either;
//# sourceMappingURL=validators.js.map